package com.dreamcodex.ti;

import com.dreamcodex.ti.component.MapCanvas;
import com.dreamcodex.ti.component.MapEditor;
import com.dreamcodex.ti.util.ECMPalette;
import com.dreamcodex.ti.util.Globals;
import com.dreamcodex.ti.util.TIGlobals;
import com.dreamcodex.ti.util.TransChar;

import javax.imageio.ImageIO;
import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.*;

/**
 * Created with IntelliJ IDEA.
 * User: RasmusM
 * Date: 02-11-13
 * Time: 17:22
 * <p/>
 * All file import/export methods moved here to shorten the main file.
 */
public class MagellanImportExport {

    public static final byte[] BIN_HEADER_MAG = {(byte) 'M', (byte) 'G'};
    public static final byte[] BIN_HEADER_VER = {(byte) '0', (byte) '1'};

    public static final byte BIN_CHUNK_COLORS = 1 << 1;
    public static final byte BIN_CHUNK_CHARS = 1 << 2;
    public static final byte BIN_CHUNK_SPRITES = 1 << 3;

    public static final byte BIN_MAP_HEADER_RESERVED1 = 0;
    public static final byte BIN_MAP_HEADER_RESERVED2 = 0;
    public static final byte BIN_MAP_HEADER_RESERVED3 = 0;
    public static final byte BIN_MAP_HEADER_RESERVED4 = 0;

    private final int ASM_LINELEN = 40;

    private MapEditor mapdMain;
    private int[][] clrSets;
    private ECMPalette[] ecmPalettes;
    private HashMap<Integer, int[][]> hmCharGrids;
    private HashMap<Integer, int[][]> hmCharColors;
    private ECMPalette[] ecmCharPalettes;
    private int colorMode;

    MagellanImportExport(
            MapEditor mapdMain,
            ECMPalette[] ecmPalettes, int[][] clrSets,
            HashMap<Integer, int[][]> hmCharGrids,
            HashMap<Integer, int[][]> hmCharColors,
            ECMPalette[] ecmCharPalettes,
            int colorMode
    ) {
        this.mapdMain = mapdMain;
        this.clrSets = clrSets;
        this.hmCharGrids = hmCharGrids;
        this.hmCharColors = hmCharColors;
        this.ecmPalettes = ecmPalettes;
        this.ecmCharPalettes = ecmCharPalettes;
        this.colorMode = colorMode;
    }

    protected int readDataFile(File mapDataFile) {
        hmCharGrids.clear();
        try {
            BufferedReader br = new BufferedReader(new FileReader(mapDataFile));
            String lineIn = "";
            int mapY = 0;
            int mapX = 0;
            int mapCount = 1;
            int mapColor = 15;
            int mapWidth = 32;
            int mapHeight = 24;
            int charStart = TIGlobals.FIRSTCHAR;
            int charEnd = TIGlobals.LASTCHAR;
            int charRead = charStart;
            int charRead1 = charStart;
            int charRead2 = charStart;
            int charColorRead = charStart;
            int cset = (int) (Math.floor(charStart / 8));
            int palNo = 0;
            int charPalNo = charStart;
            int currMap = 0;
            do {
                lineIn = br.readLine();
                if (lineIn == null) {
                    break;
                }
                else {
                    if (lineIn.startsWith(Globals.KEY_COLORSET)) {
                        lineIn = lineIn.substring(Globals.KEY_COLORSET.length());
                        clrSets[cset][Globals.INDEX_CLR_FORE] = Integer.parseInt(lineIn.substring(0, lineIn.indexOf("|")));
                        clrSets[cset][Globals.INDEX_CLR_BACK] = Integer.parseInt(lineIn.substring(lineIn.indexOf("|") + 1));
                        cset++;
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARDATA)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARDATA.length());
                        if (lineIn.length() > 16) {
                            lineIn = lineIn.substring(0, 16);
                        }
                        else if (lineIn.length() < 16) {
                            lineIn = lineIn + "0000000000000000";
                            lineIn = lineIn.substring(0, 16);
                        }
                        hmCharGrids.put(charRead, Globals.getByteIntArray(lineIn, 8));
                        charRead++;
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARDATA1) && (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARDATA1.length());
                        if (lineIn.length() > 16) {
                            lineIn = lineIn.substring(0, 16);
                        }
                        else if (lineIn.length() < 16) {
                            lineIn = lineIn + "0000000000000000";
                            lineIn = lineIn.substring(0, 16);
                        }
                        int[][] charGrid = hmCharGrids.get(charRead1++);
                        int[][] charGrid1 = Globals.getByteIntArray(lineIn, 8);
                        for (int y = 0; y < charGrid1.length; y++) {
                            for (int x = 0; x < charGrid1[0].length; x++) {
                                charGrid[y][x] |= charGrid1[y][x] << 1;
                            }
                        }
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARDATA2) && (colorMode == Magellan.COLOR_MODE_ECM_3)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARDATA2.length());
                        if (lineIn.length() > 16) {
                            lineIn = lineIn.substring(0, 16);
                        }
                        else if (lineIn.length() < 16) {
                            lineIn = lineIn + "0000000000000000";
                            lineIn = lineIn.substring(0, 16);
                        }
                        int[][] charGrid = hmCharGrids.get(charRead2++);
                        int[][] charGrid2 = Globals.getByteIntArray(lineIn, 8);
                        for (int y = 0; y < charGrid2.length; y++) {
                            for (int x = 0; x < charGrid2[0].length; x++) {
                                charGrid[y][x] |= charGrid2[y][x] << 2;
                            }
                        }
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARCOLOR) && colorMode == Magellan.COLOR_MODE_BITMAP) {
                        lineIn = lineIn.substring(Globals.KEY_CHARCOLOR.length());
                        if (lineIn.length() > 16) {
                            lineIn = lineIn.substring(0, 16);
                        }
                        else if (lineIn.length() < 16) {
                            lineIn = lineIn + "1010101010101010";
                            lineIn = lineIn.substring(0, 16);
                        }
                        int[][] charColors = new int[8][2];
                        for (int y = 0; y < 8; y++) {
                            charColors[y][1] = Integer.parseInt(Character.toString(lineIn.charAt(y * 2)), 16);
                            charColors[y][0] = Integer.parseInt(Character.toString(lineIn.charAt(y * 2 + 1)), 16);
                        }
                        hmCharColors.put(charColorRead, charColors);
                        charColorRead++;
                    }
                    else if (lineIn.startsWith(Globals.KEY_PALETTE) && (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3)) {
                        lineIn = lineIn.substring(Globals.KEY_PALETTE.length());
                        ECMPalette ecmPalette = ecmPalettes[palNo++];
                        String[] hexCols = lineIn.split("\\|");
                        for (int i = 0; i < Math.min(hexCols.length, ecmPalette.getSize()); i++) {
                            Color color;
                            try {
                                color = new Color(Integer.parseInt(hexCols[i].substring(2), 16));
                            } catch (NumberFormatException e) {
                                System.out.println(hexCols[i]);
                                color = TIGlobals.TI_COLOR_TRANSOPAQUE;
                            }
                            ecmPalette.setColor(i, color);
                        }
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARPALS) && (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3)) {
                        ECMPalette ecmPalette;
                        try {
                            ecmPalette = ecmPalettes[Integer.parseInt(lineIn.substring(Globals.KEY_CHARPALS.length()))];
                        } catch (NumberFormatException e) {
                            ecmPalette = ecmPalettes[0];
                        }
                        ecmCharPalettes[charPalNo++] = ecmPalette;
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARRANG)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARRANG.length());
                        charStart = Integer.parseInt(lineIn.substring(0, lineIn.indexOf("|")));
                        charEnd = Integer.parseInt(lineIn.substring(lineIn.indexOf("|") + 1));
                        charRead = charStart;
                        charRead1 = charStart;
                        charRead2 = charStart;
                        charColorRead = charStart;
                        cset = (int) (Math.floor(charStart / 8));
                        charPalNo = charStart;
                    }
                    else if (lineIn.startsWith(Globals.KEY_SCRBACK)) {
                        lineIn = lineIn.substring(Globals.KEY_SCRBACK.length());
                        mapColor = Integer.parseInt(lineIn);
                        mapdMain.setColorScreen(mapColor);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPCOUNT)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPCOUNT.length());
                        mapCount = Integer.parseInt(lineIn);
                    }
                    else if (lineIn.equals(Globals.KEY_MAPSTART)) {
                        if (mapY > 0) {
                            mapdMain.addBlankMap(mapWidth, mapHeight);
                            currMap++;
                            mapdMain.setCurrentMapId(currMap);
                            mapY = 0;
                        }
                    }
                    else if (lineIn.equals(Globals.KEY_MAPEND)) {
                        mapdMain.setColorScreen(mapColor);
                        mapdMain.storeCurrentMap();
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPSIZE)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPSIZE.length());
                        mapWidth = Integer.parseInt(lineIn.substring(0, lineIn.indexOf("|")));
                        mapHeight = Integer.parseInt(lineIn.substring(lineIn.indexOf("|") + 1));
                        mapdMain.setGridWidth(mapWidth);
                        mapdMain.setGridHeight(mapHeight);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPBACK)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPBACK.length());
                        mapColor = Integer.parseInt(lineIn);
                        mapdMain.setColorScreen(mapColor);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPDATA)) {
                        if (mapY >= mapHeight) {
                            mapdMain.setColorScreen(mapColor);
                            mapdMain.storeCurrentMap();
                            mapdMain.addBlankMap(mapWidth, mapHeight);
                            currMap++;
                            mapdMain.setCurrentMapId(currMap);
                            mapY = 0;
                        }
                        lineIn = lineIn.substring(Globals.KEY_MAPDATA.length());
                        StringTokenizer stParse = new StringTokenizer(lineIn, "|", false);
                        while (stParse.hasMoreTokens()) {
                            String sVal = stParse.nextToken();
                            mapdMain.setGridAt(mapX, mapY, Integer.parseInt(sVal));
                            mapX++;
                        }
                        mapX = 0;
                        mapY++;
                    }
                }
            } while (lineIn != null);
            br.close();
            if (colorMode == Magellan.COLOR_MODE_BITMAP && charColorRead == charStart) {
                // Bitmap color mode but no bitmap colors found - use color sets
                for (int i = charStart; i <= charEnd; i++) {
                    if (hmCharGrids.get(i) != null) {
                        int[][] colorGrid = hmCharColors.get(i);
                        if (colorGrid == null) {
                            colorGrid = new int[8][2];
                            hmCharColors.put(i, colorGrid);
                        }
                        for (int row = 0; row < 8; row++) {
                            colorGrid[row][0] = clrSets[i / 8][Globals.INDEX_CLR_BACK];
                            colorGrid[row][1] = clrSets[i / 8][Globals.INDEX_CLR_FORE];
                        }
                    }
                }
            }
            return charStart;
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
        return TIGlobals.MINCHAR;
    }

    protected void readAppendDataFile(File mapDataFile) {
        try {
            BufferedReader br = new BufferedReader(new FileReader(mapDataFile));
            String lineIn = "";
            int mapY = 0;
            int mapX = 0;
            int mapCount = 1;
            int mapColor = 15;
            int mapWidth = 32;
            int mapHeight = 24;
            int currMap = mapdMain.getMapCount();
            do {
                lineIn = br.readLine();
                if (lineIn == null) {
                    break;
                }
                else {
                    if (lineIn.startsWith(Globals.KEY_COLORSET)) {
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARDATA)) {
                    }
                    else if (lineIn.startsWith(Globals.KEY_SCRBACK)) {
                        lineIn = lineIn.substring(Globals.KEY_SCRBACK.length());
                        mapColor = Integer.parseInt(lineIn);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPCOUNT)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPCOUNT.length());
                        mapCount = Integer.parseInt(lineIn);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPSIZE)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPSIZE.length());
                        mapWidth = Integer.parseInt(lineIn.substring(0, lineIn.indexOf("|")));
                        mapHeight = Integer.parseInt(lineIn.substring(lineIn.indexOf("|") + 1));
                    }
                    else if (lineIn.equals(Globals.KEY_MAPSTART)) {
                        if (mapY > 0) {
                            mapdMain.addBlankMap(mapWidth, mapHeight);
                            currMap++;
                            mapdMain.setCurrentMapId(currMap);
                            mapY = 0;
                        }
                    }
                    else if (lineIn.equals(Globals.KEY_MAPEND)) {
                        mapdMain.setColorScreen(mapColor);
                        mapdMain.storeCurrentMap();
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPBACK)) {
                        lineIn = lineIn.substring(Globals.KEY_MAPBACK.length());
                        mapColor = Integer.parseInt(lineIn);
                        mapdMain.setColorScreen(mapColor);
                    }
                    else if (lineIn.startsWith(Globals.KEY_MAPDATA)) {
                        if (currMap == mapdMain.getMapCount()) {
                            mapdMain.storeCurrentMap();
                            mapdMain.addBlankMap(mapWidth, mapHeight);
                            mapdMain.setCurrentMapId(currMap);
                            mapY = 0;
                            mapdMain.setColorScreen(mapColor);
                        }
                        else if (mapY >= mapHeight) {
                            mapdMain.storeCurrentMap();
                            mapdMain.addBlankMap();
                            currMap++;
                            mapdMain.setCurrentMapId(currMap);
                            mapY = 0;
                            mapdMain.setColorScreen(mapColor);
                        }
                        lineIn = lineIn.substring(Globals.KEY_MAPDATA.length());
                        StringTokenizer stParse = new StringTokenizer(lineIn, "|", false);
                        while (stParse.hasMoreTokens()) {
                            String sVal = stParse.nextToken();
                            mapdMain.setGridAt(mapX, mapY, Integer.parseInt(sVal));
                            mapX++;
                        }
                        mapX = 0;
                        mapY++;
                    }
                }
            } while (lineIn != null);
            br.close();
            mapdMain.updateComponents();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void readCharacterData(File mapDataFile) {
        try {
            BufferedReader br = new BufferedReader(new FileReader(mapDataFile));
            String lineIn = "";
            int charStart = TIGlobals.FIRSTCHAR;
            int charEnd = TIGlobals.LASTCHAR;
            int charRead = charStart;
            do {
                lineIn = br.readLine();
                if (lineIn == null) {
                    break;
                }
                else {
                    if (lineIn.startsWith(Globals.KEY_CHARDATA)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARDATA.length());
                        hmCharGrids.put(charRead, Globals.getByteIntArray(lineIn, 8));
                        charRead++;
                    }
                    else if (lineIn.startsWith(Globals.KEY_CHARRANG)) {
                        lineIn = lineIn.substring(Globals.KEY_CHARRANG.length());
                        charStart = Integer.parseInt(lineIn.substring(0, lineIn.indexOf("|")));
                        charEnd = Integer.parseInt(lineIn.substring(lineIn.indexOf("|") + 1));
                        charRead = charStart;
                    }
                }
            } while (lineIn != null);
            br.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void readVramDumpFile(File vramDumpFile, int charTableOffset, int mapTableOffset, int colorTableOffset, boolean bitmapMode) {
        try {
            FileInputStream fib = new FileInputStream(vramDumpFile);
            boolean basicOffset = false;
            if (charTableOffset == mapTableOffset)
                basicOffset = true;
            int readPos = 0;
            int readInt;
            int charStart = TIGlobals.MINCHAR;
            int charEnd = TIGlobals.MAXCHAR;
            StringBuffer sbChar = new StringBuffer();
            int charByte = 0;
            int charRead = charStart;
            int charTableLength = 256;
            int colorTableLength = bitmapMode ? 2048 : 32;
            int mapCols = 32;
            int mapRows = 24;
            int mapTableLength = (mapCols * mapRows);
            while ((readInt = fib.read()) != -1) {
                if ((readPos >= charTableOffset) && (readPos < (charTableOffset + (charTableLength * 8)))) {
                    if (charRead <= charEnd) {
                        sbChar.append((readInt < 16 ? "0" : "") + Integer.toHexString(readInt));
                        charByte++;
                        if (charByte >= 8) {
                            int mapChar = charRead;
                            if (basicOffset)
                                mapChar = (mapChar + 0xA0) & 0xFF;
                            hmCharGrids.put(mapChar, Globals.getByteIntArray(sbChar.toString(), 8));
                            charRead++;
                            charByte = 0;
                            sbChar.delete(0, sbChar.length());
                        }
                    }
                }
                if (readPos >= mapTableOffset && readPos < (mapTableOffset + mapTableLength)) {
                    int mapCell = readPos - mapTableOffset;
                    int mapRow = (int) (Math.floor(mapCell / mapCols));
                    int mapCol = mapCell % mapCols;
                    int mapChar = readInt;
                    if (basicOffset)
                        mapChar = (mapChar + 0xA0) & 0xFF;
                    mapdMain.setGridAt(mapCol, mapRow, mapChar);
                }
                if (readPos >= colorTableOffset && readPos < (colorTableOffset + colorTableLength)) {
                    if (bitmapMode) {
                        if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                            int colorByte = readPos - colorTableOffset;
                            int colorChar = colorByte / 8;
                            int[][] colorGrid = hmCharColors.get(colorChar);
                            if (colorGrid == null) {
                                colorGrid = new int[8][2];
                                hmCharColors.put(colorChar, colorGrid);
                            }
                            int row = colorByte % 8;
                            colorGrid[row][0] = readInt & 0x0F;
                            colorGrid[row][1] = (readInt & 0xF0) >> 4;
                        }
                    }
                    else {
                        int setNum = readPos - colorTableOffset;
                        int colorFore = (readInt & 0xF0) >> 4;
                        int colorBack = (readInt & 0x0F);
                        if (basicOffset)
                            setNum = (setNum + 20) & 0x1F;
                        clrSets[setNum][Globals.INDEX_CLR_FORE] = colorFore;
                        clrSets[setNum][Globals.INDEX_CLR_BACK] = colorBack;
                        if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                            for (int colorChar = setNum * 8; colorChar < setNum * 8 + 8; colorChar++) {
                                int[][] colorGrid = hmCharColors.get(colorChar);
                                if (colorGrid == null) {
                                    colorGrid = new int[8][2];
                                    hmCharColors.put(colorChar, colorGrid);
                                }
                                for (int row = 0; row < 8; row++) {
                                    colorGrid[row][0] = colorBack;
                                    colorGrid[row][1] = colorFore;
                                }
                            }
                        }
                    }
                }
                readPos++;
            }
            fib.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void writeDataFile(File mapDataFile) {
        try {
            // store working map first
            mapdMain.storeCurrentMap();
            // get file output buffer
            BufferedWriter bw = new BufferedWriter(new FileWriter(mapDataFile));
            // output overall character range (this is for backwards compatibility with earlier Magellan releases, will be phased out)
            bw.write("* CHARACTER RANGE");
            bw.newLine();
            bw.write(Globals.KEY_CHARRANG + TIGlobals.MINCHAR + "|" + +TIGlobals.MAXCHAR);
            bw.newLine();
            // save colorsets
            bw.write("* COLORSET");
            bw.newLine();
            for (int i = 0; i < clrSets.length; i++) {
                bw.write(Globals.KEY_COLORSET + clrSets[i][Globals.INDEX_CLR_FORE] + "|" + clrSets[i][Globals.INDEX_CLR_BACK]);
                bw.newLine();
            }
            // save palettes
            if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                bw.write("* PALETTES");
                bw.newLine();
                Map<ECMPalette, Integer> paletteMap = new HashMap<ECMPalette, Integer>(256);
                int n = 0;
                for (ECMPalette ecmPalette : ecmPalettes) {
                    bw.write(Globals.KEY_PALETTE);
                    Color[] colors = ecmPalette.getColors();
                    for (int i = 0; i < colors.length; i++) {
                        bw.write(Integer.toHexString(colors[i].getRGB()));
                        if (i < colors.length - 1) {
                            bw.write("|");
                        }
                    }
                    bw.newLine();
                    paletteMap.put(ecmPalette, n++);
                }
                // save character palette numbers
                bw.write("* CHAR PALETTE NUMBERS");
                bw.newLine();
                for (ECMPalette charPalette : ecmCharPalettes) {
                    bw.write(Globals.KEY_CHARPALS);
                    bw.write(Integer.toString(paletteMap.get(charPalette)));
                    bw.newLine();
                }
            }
            // save chardefs
            bw.write("* CHAR DEFS");
            bw.newLine();
            for (int i = TIGlobals.MINCHAR; i <= TIGlobals.MAXCHAR; i++) {
                bw.write(Globals.KEY_CHARDATA);
                if (hmCharGrids.get(i) != null) {
                    String hexstr = Globals.getByteString(hmCharGrids.get(i));
                    bw.write(hexstr, 0, hexstr.length());
                }
                else {
                    bw.write(Globals.BLANKCHAR, 0, Globals.BLANKCHAR.length());
                }
                bw.newLine();
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                bw.write("* CHAR DEFS PLANE 1");
                bw.newLine();
                for (int i = TIGlobals.MINCHAR; i <= TIGlobals.MAXCHAR; i++) {
                    bw.write(Globals.KEY_CHARDATA1);
                    if (hmCharGrids.get(i) != null) {
                        String hexstr = Globals.getByteString(hmCharGrids.get(i), 2);
                        bw.write(hexstr, 0, hexstr.length());
                    }
                    else {
                        bw.write(Globals.BLANKCHAR, 0, Globals.BLANKCHAR.length());
                    }
                    bw.newLine();
                }
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_3) {
                bw.write("* CHAR DEFS PLANE 2");
                bw.newLine();
                for (int i = TIGlobals.MINCHAR; i <= TIGlobals.MAXCHAR; i++) {
                    bw.write(Globals.KEY_CHARDATA2);
                    if (hmCharGrids.get(i) != null) {
                        String hexstr = Globals.getByteString(hmCharGrids.get(i), 4);
                        bw.write(hexstr, 0, hexstr.length());
                    }
                    else {
                        bw.write(Globals.BLANKCHAR, 0, Globals.BLANKCHAR.length());
                    }
                    bw.newLine();
                }
            }
            // Save char colors (bitmap mode)
            if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                bw.write("* CHAR COLORS");
                bw.newLine();
                for (int i = TIGlobals.MINCHAR; i <= TIGlobals.MAXCHAR; i++) {
                    bw.write(Globals.KEY_CHARCOLOR);
                    int[][] charColors = hmCharColors.get(i);
                    if (charColors != null) {
                        String hexstr = "";
                        for (int[] rowColors : charColors) {
                            hexstr += Integer.toHexString(rowColors[1]);
                            hexstr += Integer.toHexString(rowColors[0]);
                        }
                        bw.write(hexstr, 0, hexstr.length());
                    }
                    else {
                        bw.write(Globals.BLANKCHAR, 0, Globals.BLANKCHAR.length());
                    }
                    bw.newLine();
                }
            }
            // save map parameters
            bw.write("* MAPS");
            bw.newLine();
            bw.write(Globals.KEY_MAPCOUNT + mapdMain.getMapCount());
            bw.newLine();
            // save map(s)
            for (int m = 0; m < mapdMain.getMapCount(); m++) {
                bw.write("* MAP #" + (m + 1));
                bw.newLine();
                bw.write(Globals.KEY_MAPSTART);
                bw.newLine();
                int[][] mapToSave = mapdMain.getMapData(m);
                bw.write(Globals.KEY_MAPSIZE + mapToSave[0].length + "|" + mapToSave.length);
                bw.newLine();
                bw.write(Globals.KEY_MAPBACK + mapdMain.getScreenColor(m));
                bw.newLine();
                for (int y = 0; y < mapToSave.length; y++) {
                    bw.write(Globals.KEY_MAPDATA);
                    for (int x = 0; x < mapToSave[y].length; x++) {
                        bw.write((x > 0 ? "|" : "") + mapToSave[y][x]);
                    }
                    bw.newLine();
                }
                bw.write(Globals.KEY_MAPEND);
                bw.newLine();
            }
            bw.flush();
            bw.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void writeXBDataFile(File mapDataFile, int startChar, int endChar, int codeLine, int charLine, int mapLine, int interLine, boolean addExecutableCode, boolean includeComments, boolean currMapOnly) {
        int currLine = codeLine;
        int itemCount = 0;
        int colorDataStart = 0;
        int colorSetStart = (int) (Math.floor(startChar / 8));
        int colorSetEnd = (int) (Math.floor(endChar / 8));
        int colorSetNum = (int) (Math.floor((startChar - TIGlobals.FIRSTCHAR) / 8)) + 1;
        int colorCount = (colorSetEnd - colorSetStart) + 1;
        int charDataStart;
        int charCount = 0;
        int[] mapDataStart = new int[mapdMain.getMapCount()];
        int mapCols;
        int mapRows;
        StringBuffer sbOutLine = new StringBuffer();
        try {
            mapdMain.storeCurrentMap();
            BufferedWriter bw = new BufferedWriter(new FileWriter(mapDataFile));
            colorDataStart = currLine;
            if (includeComments) {
                bw.write("REM COLORSET DECLARATIONS");
                bw.newLine();
                bw.newLine();
            }
            int blockCount = 0;
            bw.write(currLine + " DATA ");
            for (int i = colorSetStart; i < (colorSetStart + colorCount); i++) {
                if (blockCount > 7) {
                    bw.newLine();
                    currLine = currLine + interLine;
                    bw.write(currLine + " DATA ");
                    blockCount = 0;
                }
                if (blockCount > 0) {
                    bw.write(",");
                }
                bw.write(colorSetNum + "," + (clrSets[i][Globals.INDEX_CLR_FORE] + 1) + "," + (clrSets[i][Globals.INDEX_CLR_BACK] + 1));
                colorSetNum++;
                blockCount++;
            }
            bw.newLine();
            currLine = currLine + interLine;
            if (includeComments) {
                bw.newLine();
                bw.write("REM CHARACTER DATA");
                bw.newLine();
                bw.newLine();
            }
            currLine = charLine;
            charDataStart = currLine;
            for (int i = startChar; i <= endChar; i++) {
                String hexstr;
                if (hmCharGrids.get(i) != null) {
                    hexstr = Globals.getByteString(hmCharGrids.get(i));
                }
                else {
                    hexstr = Globals.BLANKCHAR;
                }
                if (!hexstr.equals(Globals.BLANKCHAR)) {
                    if (itemCount == 0) {
                        sbOutLine.append(currLine + " DATA ");
                    }
                    else {
                        sbOutLine.append(",");
                    }
                    sbOutLine.append(i + "," + '"' + hexstr.toUpperCase() + '"');
                    itemCount++;
                    if (itemCount >= 4) {
                        bw.write(sbOutLine.toString());
                        bw.newLine();
                        sbOutLine.delete(0, sbOutLine.length());
                        currLine = currLine + interLine;
                        itemCount = 0;
                    }
                    charCount++;
                }
            }
            if (sbOutLine.length() > 0) {
                bw.write(sbOutLine.toString());
                bw.newLine();
            }
            currLine = currLine + interLine;
            itemCount = 0;
            sbOutLine.delete(0, sbOutLine.length());
            currLine = mapLine;
            if (includeComments) {
                bw.newLine();
                bw.write("REM MAP DATA");
                bw.newLine();
            }
            int mapTicker = 1;
            for (int m = 0; m < mapdMain.getMapCount(); m++) {
                if (!currMapOnly || m == mapdMain.getCurrentMapId()) {
                    int[][] mapToSave = mapdMain.getMapData(m);
                    mapCols = mapToSave[0].length;
                    mapRows = mapToSave.length;
                    mapDataStart[m] = currLine;
                    if (includeComments) {
                        bw.newLine();
                        bw.write("REM MAP #" + mapTicker);
                        bw.newLine();
                    }
                    int mapColor = mapdMain.getScreenColorTI(m);
                    if (includeComments) {
                        bw.write("REM MAP #" + mapTicker + " WIDTH, HEIGHT, SCREEN COLOR");
                        bw.newLine();
                    }
                    bw.write(currLine + " DATA " + mapCols + "," + mapRows + "," + mapColor);
                    bw.newLine();
                    if (includeComments) {
                        bw.write("REM MAP #" + mapTicker + " DATA");
                        bw.newLine();
                    }
                    currLine = currLine + interLine;
                    int mapTileVal = TIGlobals.SPACECHAR;
                    for (int y = 0; y < mapRows; y++) {
                        for (int x = 0; x < mapCols; x++) {
                            if (itemCount == 0) {
                                sbOutLine.append(currLine + " DATA ");
                            }
                            else {
                                sbOutLine.append(",");
                            }
                            mapTileVal = ((mapToSave[y][x] < startChar || mapToSave[y][x] > endChar) ? TIGlobals.SPACECHAR : mapToSave[y][x]);
                            sbOutLine.append("" + mapTileVal);
                            itemCount++;
                            if (itemCount >= 16) {
                                bw.write(sbOutLine.toString());
                                bw.newLine();
                                sbOutLine.delete(0, sbOutLine.length());
                                currLine = currLine + interLine;
                                itemCount = 0;
                            }
                        }
                    }
                    mapTicker++;
                }
            }
            if (addExecutableCode) {
                if (includeComments) {
                    bw.newLine();
                    bw.write("REM LOAD COLORSET");
                    bw.newLine();
                    bw.newLine();
                }
                bw.write(currLine + " RESTORE " + colorDataStart + "::FOR C=1 TO " + colorCount + "::READ CS,CF,CB::CALL COLOR(CS,CF,CB)::NEXT C");
                currLine = currLine + interLine;
                bw.newLine();
                if (includeComments) {
                    bw.newLine();
                    bw.write("REM LOAD CHARACTERS");
                    bw.newLine();
                    bw.newLine();
                }
                bw.write(currLine + " RESTORE " + charDataStart + "::FOR C=1 TO " + charCount + "::READ CN,CC$::CALL CHAR(CN,CC$)::NEXT C");
                currLine = currLine + interLine;
                bw.newLine();
                if (includeComments) {
                    bw.newLine();
                    bw.write("REM DRAW MAP(S)");
                    bw.newLine();
                    bw.newLine();
                }
                for (int mp = 0; mp < mapdMain.getMapCount(); mp++) {
                    if (!currMapOnly || mp == mapdMain.getCurrentMapId()) {
                        bw.write(currLine + " CALL CLEAR");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " RESTORE " + mapDataStart[mp]);
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " READ W,H,SC::CALL SCREEN(SC)::CALL CLEAR");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " FOR Y=1 TO H");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " FOR X=1 TO W");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " READ CP::CALL VCHAR(Y,X,CP)");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " NEXT X");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " NEXT Y");
                        currLine = currLine + interLine;
                        bw.newLine();
                        bw.write(currLine + " CALL KEY(0,K,S)::IF S=0 THEN " + currLine);
                        currLine = currLine + interLine;
                        bw.newLine();
                    }
                }
                bw.write(currLine + " END");
                currLine = currLine + interLine;
                bw.newLine();
            }
            bw.flush();
            bw.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void writeASMDataFile(File mapDataFile, int startChar, int endChar, boolean includeComments, boolean currMapOnly, boolean includeCharNumbers) {
        int itemCount = 0;
        StringBuffer sbLine = new StringBuffer();
        try {
            mapdMain.storeCurrentMap();
            BufferedWriter bw = new BufferedWriter(new FileWriter(mapDataFile));
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Colorset Definitions", false);
                printPaddedLine(bw, "****************************************", false);
            }
            if (colorMode == Magellan.COLOR_MODE_GRAPHICS_1) {
                printPaddedLine(bw, "CLRNUM DATA " + (clrSets.length - 1), includeComments);
                for (int i = 0; i < clrSets.length; i++) {
                    if (itemCount == 0) {
                        if (i == 0) {
                            sbLine.append("CLRSET BYTE ");
                        }
                        else {
                            sbLine.append("       BYTE ");
                        }
                    }
                    if (itemCount > 0) {
                        sbLine.append(",");
                    }
                    sbLine.append(">");
                    sbLine.append(Integer.toHexString(clrSets[i][Globals.INDEX_CLR_FORE]).toUpperCase());
                    sbLine.append(Integer.toHexString(clrSets[i][Globals.INDEX_CLR_BACK]).toUpperCase());
                    itemCount++;
                    if (itemCount > 3) {
                        printPaddedLine(bw, sbLine.toString(), includeComments);
                        sbLine.delete(0, sbLine.length());
                        itemCount = 0;
                    }
                }
            }
            else if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                sbLine.delete(0, sbLine.length());
                itemCount = 0;
                for (int i = startChar; i <= endChar; i++) {
                    int[][] charColors = hmCharColors.get(i);
                    if (charColors != null && !Globals.isColorGridEmpty(charColors)) {
                        if (includeCharNumbers) {
                            printPaddedLine(bw, "CCH" + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA " + i, includeComments);
                        }
                        sbLine.append("COL").append(itemCount).append(itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")).append(" DATA ");
                        for (int row = 0; row < 8; row += 2) {
                            sbLine.append(">");
                            int[] rowColors = charColors[row];
                            sbLine.append(Integer.toHexString(rowColors[1]).toUpperCase());
                            sbLine.append(Integer.toHexString(rowColors[0]).toUpperCase());
                            rowColors = charColors[row + 1];
                            sbLine.append(Integer.toHexString(rowColors[1]).toUpperCase());
                            sbLine.append(Integer.toHexString(rowColors[0]).toUpperCase());
                            if (row < 6) {
                                sbLine.append(",");
                            }
                        }
                        printPaddedLine(bw, sbLine.toString(), includeComments);
                        sbLine.delete(0, sbLine.length());
                        itemCount++;
                    }
                }
            }
            else if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                sbLine.delete(0, sbLine.length());
                Map<ECMPalette, Integer> paletteMap = new HashMap<ECMPalette, Integer>(256);
                for (int i = 0; i < ecmPalettes.length; i++) {
                    ECMPalette ecmPalette = ecmPalettes[i];
                    for (int l = 0; l < ecmPalette.getSize() / 4; l++) {
                        if (l == 0) {
                            sbLine.append("PAL").append(i).append(i < 10 ? " " : "").append("  DATA ");
                        }
                        else {
                            sbLine.append("       DATA ");
                        }
                        for (int c = 0; c < 4; c++) {
                            Color color = ecmPalette.getColor(c + l * 4);
                            sbLine.append(">0");
                            sbLine.append(Integer.toHexString(color.getRed() >> 4).toUpperCase());
                            sbLine.append(Integer.toHexString(color.getGreen() >> 4).toUpperCase());
                            sbLine.append(Integer.toHexString(color.getBlue() >> 4).toUpperCase());
                            if (c < 3) {
                                sbLine.append(",");
                            }
                        }
                        printPaddedLine(bw, sbLine.toString(), includeComments);
                        sbLine.delete(0, sbLine.length());
                    }
                    paletteMap.put(ecmPalette, i);
                }
                if (includeComments) {
                    printPaddedLine(bw, "****************************************", false);
                    printPaddedLine(bw, "* Tile Attributes", false);
                    printPaddedLine(bw, "****************************************", false);
                }
                sbLine.delete(0, sbLine.length());
                for (int i = startChar; i <= endChar; i++) {
                    sbLine.append("TAT").append(i).append(i < 10 ? "   " : (i < 100 ? "  " : " "));
                    sbLine.append("BYTE >").append(Globals.toHexString(paletteMap.get(ecmCharPalettes[i]) << (colorMode == Magellan.COLOR_MODE_ECM_3 ? 1 : 0), 2));
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                    sbLine.delete(0, sbLine.length());
                }
            }
            if (sbLine.length() > 0) {
                printPaddedLine(bw, sbLine.toString(), includeComments);
                sbLine.delete(0, sbLine.length());
            }
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Character Patterns" + (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3 ? " Plane 0" : ""), false);
                printPaddedLine(bw, "****************************************", false);
            }
            sbLine.delete(0, sbLine.length());
            itemCount = 0;
            for (int i = startChar; i <= endChar; i++) {
                String hexstr = Globals.BLANKCHAR;
                if (hmCharGrids.get(i) != null) {
                    hexstr = Globals.getByteString(hmCharGrids.get(i)).toUpperCase();
                }
                // Blank chars are not exported if the background is transparent
                if (!hexstr.equals(Globals.BLANKCHAR) || colorMode == Magellan.COLOR_MODE_GRAPHICS_1 && clrSets[(int) (Math.floor(i / 8))][Globals.INDEX_CLR_BACK] != 0 || colorMode == Magellan.COLOR_MODE_BITMAP && !Globals.isColorGridEmpty(hmCharColors.get(i)) || colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                    if (includeCharNumbers) {
                        printPaddedLine(bw, "PCH" + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA " + i, includeComments);
                    }
                    sbLine.append((colorMode == Magellan.COLOR_MODE_GRAPHICS_1 || colorMode == Magellan.COLOR_MODE_BITMAP ? "PAT" : "P0_") + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA ");
                    sbLine.append(">" + hexstr.substring(0, 4) + ",");
                    sbLine.append(">" + hexstr.substring(4, 8) + ",");
                    sbLine.append(">" + hexstr.substring(8, 12) + ",");
                    sbLine.append(">" + hexstr.substring(12, 16));
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                    sbLine.delete(0, sbLine.length());
                    itemCount++;
                }
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                if (includeComments) {
                    printPaddedLine(bw, "****************************************", false);
                    printPaddedLine(bw, "* Character Patterns Plane 1", false);
                    printPaddedLine(bw, "****************************************", false);
                }
                itemCount = 0;
                for (int i = startChar; i <= endChar; i++) {
                    String hexstr = Globals.BLANKCHAR;
                    if (hmCharGrids.get(i) != null) {
                        hexstr = Globals.getByteString(hmCharGrids.get(i), 2).toUpperCase();
                    }
                    sbLine.append("P1_" + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA ");
                    sbLine.append(">" + hexstr.substring(0, 4) + ",");
                    sbLine.append(">" + hexstr.substring(4, 8) + ",");
                    sbLine.append(">" + hexstr.substring(8, 12) + ",");
                    sbLine.append(">" + hexstr.substring(12, 16));
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                    sbLine.delete(0, sbLine.length());
                    itemCount++;
                }
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_3) {
                if (includeComments) {
                    printPaddedLine(bw, "****************************************", false);
                    printPaddedLine(bw, "* Character Patterns Plane 2", false);
                    printPaddedLine(bw, "****************************************", false);
                }
                itemCount = 0;
                for (int i = startChar; i <= endChar; i++) {
                    String hexstr = Globals.BLANKCHAR;
                    if (hmCharGrids.get(i) != null) {
                        hexstr = Globals.getByteString(hmCharGrids.get(i), 4).toUpperCase();
                    }
                    sbLine.append("P2_" + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA ");
                    sbLine.append(">" + hexstr.substring(0, 4) + ",");
                    sbLine.append(">" + hexstr.substring(4, 8) + ",");
                    sbLine.append(">" + hexstr.substring(8, 12) + ",");
                    sbLine.append(">" + hexstr.substring(12, 16));
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                    sbLine.delete(0, sbLine.length());
                    itemCount++;
                }
            }
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Map Data", false);
                printPaddedLine(bw, "****************************************", false);
            }
            printPaddedLine(bw, "MCOUNT DATA " + (currMapOnly ? 1 : mapdMain.getMapCount()), includeComments);
            sbLine.delete(0, sbLine.length());
            boolean isFirstByte;
            String hexChunk;
            for (int m = 0; m < mapdMain.getMapCount(); m++) {
                if (!currMapOnly || m == mapdMain.getCurrentMapId()) {
                    int[][] mapToSave = mapdMain.getMapData(m);
                    if (includeComments) {
                        printPaddedLine(bw, "* == Map #" + m + " == ", false);
                    }
                    printPaddedLine(bw, "MC" + m + (m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))) + " DATA " + mapdMain.getScreenColor(m), includeComments);
                    sbLine.append("MS").append(m).append(m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))).append(" DATA >");
                    hexChunk = Integer.toHexString(mapToSave[0].length).toUpperCase();
                    sbLine.append(hexChunk.length() < 1 ? "00" : (hexChunk.length() < 2 ? "0" : "")).append(hexChunk);
                    hexChunk = Integer.toHexString(mapToSave.length).toUpperCase();
                    sbLine.append(hexChunk.length() < 1 ? "00" : (hexChunk.length() < 2 ? "0" : "")).append(hexChunk);
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                    sbLine.delete(0, sbLine.length());
                    isFirstByte = true;
                    for (int y = 0; y < mapToSave.length; y++) {
                        if (includeComments) {
                            printPaddedLine(bw, "* -- Map Row " + y + " -- ", false);
                        }
                        for (int cl = 0; cl < Math.ceil((double) mapToSave[y].length / 8); cl++) {
                            if (y == 0 && cl == 0) {
                                sbLine.append("MD").append(m).append(m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))).append(" DATA ");
                            }
                            else {
                                sbLine.append("       DATA ");
                            }
                            for (int colpos = (cl * 8); colpos < Math.min((cl + 1) * 8, mapToSave[y].length); colpos++) {
                                if (isFirstByte) {
                                    if (colpos > (cl * 8)) {
                                        sbLine.append(",");
                                    }
                                    sbLine.append(">");
                                }
                                hexChunk = Integer.toHexString(mapToSave[y][colpos]).toUpperCase();
                                if (mapToSave[y][colpos] == MapCanvas.NOCHAR) {
                                    hexChunk = "00";
                                }
                                sbLine.append(hexChunk.length() < 1 ? "00" : (hexChunk.length() < 2 ? "0" : "")).append(hexChunk);
                                isFirstByte = !isFirstByte;
                            }
                            if (!isFirstByte) {
                                sbLine.append("XX"); // If odd, pad with an illegal value
                                isFirstByte = !isFirstByte;
                            }
                            printPaddedLine(bw, sbLine.toString(), includeComments);
                            sbLine.delete(0, sbLine.length());
                        }
                    }
                }
            }
            bw.flush();
            bw.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void writeScrollFile(File mapDataFile, boolean vertical, boolean wrap, boolean compress, boolean compressWords, boolean includeComments, boolean currMapOnly, boolean includeCharNumbers, int frames) throws Exception {
        // store working map first
        mapdMain.storeCurrentMap();
        BufferedWriter bw = null;
        try {
            ArrayList<int[][]> transMaps = new ArrayList<int[][]>();
            Map<String, TransChar> transChars = new HashMap<String, TransChar>();
            Map<Integer, ArrayList<TransChar>> colorSets = new TreeMap<Integer, ArrayList<TransChar>>();
            TransChar[] transCharSet = new TransChar[256];
            boolean[] usedChars = new boolean[256];
            int startChar = 255;
            int endChar = 0;
            int imax = 0;
            boolean allColorsOK = true;
            for (int m = 0; m < mapdMain.getMapCount(); m++) {
                if (!currMapOnly || m == mapdMain.getCurrentMapId()) {
                    int[][] mapData = mapdMain.getMapData(m);
                    if (mapData.length > 1 && mapData[0].length > 1) {
                        int i = 0;
                        for (int y = (vertical && !wrap ? 1 : 0); y < mapData.length; y++) {
                            for (int x = 0; x < mapData[0].length - (vertical || wrap ? 0 : 1); x++) {
                                int fromChar = mapData[y][x];
                                if (fromChar < startChar) {
                                    startChar = fromChar;
                                }
                                if (fromChar > endChar) {
                                    endChar = fromChar;
                                }
                                usedChars[fromChar] = true;
                                int toChar = vertical ? mapData[y > 0 ? y - 1 : mapData.length - 1][x] : mapData[y][x < mapData[0].length - 1 ? x + 1 : 0];
                                if (toChar < startChar) {
                                    startChar = toChar;
                                }
                                if (toChar > endChar) {
                                    endChar = toChar;
                                }
                                usedChars[toChar] = true;
                                String key = fromChar + "-" + toChar;
                                TransChar transChar = transChars.get(key);
                                if (transChar != null) {
                                    transChar.incCount();
                                }
                                else {
                                    boolean colorsOK = true;
                                    if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                                        int[][] charColors = new int[8][8];
                                        if (!vertical) {
                                            int[][] fromColorGrid = hmCharColors.get(fromChar);
                                            int[][] toColorGrid = hmCharColors.get(toChar);
                                            for (int r = 0; r < 8 && colorsOK; r++) {
                                                int[] fromColorRow = fromColorGrid[r];
                                                int[] toColorRow = toColorGrid[r];
                                                int screenColor = mapdMain.getColorScreen();
                                                int fromForeColor = fromColorRow[Globals.INDEX_CLR_FORE] != 0 ? fromColorRow[Globals.INDEX_CLR_FORE] : screenColor;
                                                int toForeColor = toColorRow[Globals.INDEX_CLR_FORE] != 0 ? toColorRow[Globals.INDEX_CLR_FORE] : screenColor;
                                                if (fromForeColor == toForeColor) {
                                                    charColors[r][Globals.INDEX_CLR_FORE] = fromForeColor;
                                                }
                                                else if (!Globals.arrayContains(hmCharGrids.get(fromChar)[r], Globals.INDEX_CLR_FORE)) {
                                                    charColors[r][Globals.INDEX_CLR_FORE] = toForeColor;
                                                }
                                                else if (!Globals.arrayContains(hmCharGrids.get(toChar)[r], Globals.INDEX_CLR_FORE)) {
                                                    charColors[r][Globals.INDEX_CLR_FORE] = fromForeColor;
                                                }
                                                else {
                                                    colorsOK = false;
                                                    allColorsOK = false;
                                                }
                                                int fromBackColor = fromColorRow[Globals.INDEX_CLR_BACK] != 0 ? fromColorRow[Globals.INDEX_CLR_BACK] : screenColor;
                                                int toBackColor = toColorRow[Globals.INDEX_CLR_BACK] != 0 ? toColorRow[Globals.INDEX_CLR_BACK] : screenColor;
                                                if (fromBackColor == toBackColor) {
                                                    charColors[r][Globals.INDEX_CLR_BACK] = fromBackColor;
                                                }
                                                else if (!Globals.arrayContains(hmCharGrids.get(fromChar)[r], Globals.INDEX_CLR_BACK)) {
                                                    charColors[r][Globals.INDEX_CLR_BACK] = toBackColor;
                                                }
                                                else if (!Globals.arrayContains(hmCharGrids.get(toChar)[r], Globals.INDEX_CLR_BACK)) {
                                                    charColors[r][Globals.INDEX_CLR_BACK] = fromBackColor;
                                                }
                                                else {
                                                    colorsOK = false;
                                                    allColorsOK = false;
                                                }
                                            }
                                        }
                                        transChar = new TransChar(fromChar, toChar, i, colorsOK, charColors);
                                        transChars.put(key, transChar);
                                        transCharSet[i] = transChar;
                                        imax = i++;
                                        if (imax > 255) {
                                            throw new Exception("Character Set Full: Scrolling this map requires more than 256 characters.");
                                        }
                                    }
                                    else {
                                        int foreColor = 0;
                                        int backColor = 0;
                                        int[] fromClrSet = clrSets[fromChar / 8];
                                        int[] toClrSet = clrSets[toChar / 8];
                                        int screenColor = mapdMain.getColorScreen();
                                        int fromForeColor = fromClrSet[Globals.INDEX_CLR_FORE] != 0 ? fromClrSet[Globals.INDEX_CLR_FORE] : screenColor;
                                        int toForeColor = toClrSet[Globals.INDEX_CLR_FORE] != 0 ? toClrSet[Globals.INDEX_CLR_FORE] : screenColor;
                                        if (fromForeColor == toForeColor) {
                                            foreColor = fromForeColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(fromChar), Globals.INDEX_CLR_FORE)) {
                                            foreColor = toForeColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(toChar), Globals.INDEX_CLR_FORE)) {
                                            foreColor = fromForeColor;
                                        }
                                        else {
                                            colorsOK = false;
                                            allColorsOK = false;
                                        }
                                        int fromBackColor = fromClrSet[Globals.INDEX_CLR_BACK] != 0 ? fromClrSet[Globals.INDEX_CLR_BACK] : screenColor;
                                        int toBackColor = toClrSet[Globals.INDEX_CLR_BACK] != 0 ? toClrSet[Globals.INDEX_CLR_BACK] : screenColor;
                                        if (fromBackColor == toBackColor) {
                                            backColor = fromBackColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(fromChar), Globals.INDEX_CLR_BACK)) {
                                            backColor = toBackColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(toChar), Globals.INDEX_CLR_BACK)) {
                                            backColor = fromBackColor;
                                        }
                                        else {
                                            colorsOK = false;
                                            allColorsOK = false;
                                        }
                                        transChar = new TransChar(fromChar, toChar, -1, colorsOK, foreColor, backColor);
                                        transChars.put(key, transChar);
                                        int ckey = backColor + (foreColor << 4);
                                        ArrayList<TransChar> colorSet = colorSets.get(ckey);
                                        if (colorSet == null) {
                                            colorSet = new ArrayList<TransChar>();
                                            colorSets.put(ckey, colorSet);
                                        }
                                        colorSet.add(transChar);
                                    }
                                }
                            }
                        }
                        if (colorMode != Magellan.COLOR_MODE_BITMAP) {
                            // Organize into color sets
                            i = 0;
                            for (int ckey : colorSets.keySet()) {
                                ArrayList<TransChar> colorSet = colorSets.get(ckey);
                                for (TransChar transChar : colorSet) {
                                    transChar.setIndex(i);
                                    transCharSet[i] = transChar;
                                    imax = i++;
                                    if (i > 255) {
                                        throw new Exception("Character Set Full: Scrolling this map requires more than 32 color sets.");
                                    }
                                    if (imax > 255) {
                                        throw new Exception("Character Set Full: Scrolling this map requires more than 256 characters.");
                                    }
                                }
                                while (i % 8 != 0) {
                                    i++;
                                }
                                if (i > 255) {
                                    throw new Exception("Character Set Full: Scrolling this map requires more than 32 color sets.");
                                }
                            }
                        }
                        int[][] transMap = new int[mapData.length - (vertical && !wrap ? 1 : 0)][mapData[0].length - (vertical || wrap ? 0 : 1)];
                        transMaps.add(transMap);
                        for (int y = (vertical && !wrap ? 1 : 0); y < mapData.length; y++) {
                            for (int x = 0; x < mapData[0].length - (vertical || wrap ? 0 : 1); x++) {
                                int fromChar = mapData[y][x];
                                int toChar = vertical ? mapData[y > 0 ? y - 1 : mapData.length - 1][x] : mapData[y][x < mapData[0].length - 1 ? x + 1 : 0];
                                String key = fromChar + "-" + toChar;
                                TransChar transChar = transChars.get(key);
                                transMap[y - (vertical && !wrap ? 1 : 0)][x] = transChar.getIndex();
                            }
                        }
                    }
                }
            }
            if (!allColorsOK) {
                JOptionPane.showMessageDialog(null, "Warning - Some character transitions would require more colors than available.", "Invalid Color Transitions", JOptionPane.INFORMATION_MESSAGE);
            }
            bw = new BufferedWriter(new FileWriter(mapDataFile));
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Original Character Patterns", false);
                printPaddedLine(bw, "****************************************", false);
            }
            for (int i = startChar; i <= endChar; i++) {
                String hexstr = Globals.BLANKCHAR;
                if (hmCharGrids.get(i) != null) {
                    hexstr = Globals.getByteString(hmCharGrids.get(i)).toUpperCase();
                }
                if (includeCharNumbers) {
                    printPaddedLine(bw, "PCH" + i + (i < 10 ? "  " : (i < 100 ? " " : "")) + " DATA >" + Globals.toHexString(i, 2), includeComments);
                }
                StringBuilder sbLine = new StringBuilder();
                sbLine.append("PAT").append(i).append(i < 10 ? "  " : (i < 100 ? " " : "")).append(" DATA ");
                sbLine.append(">").append(hexstr.substring(0, 4)).append(",");
                sbLine.append(">").append(hexstr.substring(4, 8)).append(",");
                sbLine.append(">").append(hexstr.substring(8, 12)).append(",");
                sbLine.append(">").append(hexstr.substring(12, 16));
                printPaddedLine(bw, sbLine.toString(), includeComments ? (usedChars[i] ? "" : "unused") : null);
            }
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Colorset Definitions", false);
                printPaddedLine(bw, "****************************************", false);
            }
            if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                int itemCount = 0;
                for (int i = startChar; i <= endChar; i++) {
                    int[][] charColors = hmCharColors.get(i);
                    if (charColors != null && !Globals.isColorGridEmpty(charColors)) {
                        if (includeCharNumbers) {
                            printPaddedLine(bw, "CCH" + itemCount + (itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")) + " DATA " + Globals.toHexString(i, 2), includeComments);
                        }
                        StringBuilder sbLine = new StringBuilder();
                        sbLine.append("COL").append(itemCount).append(itemCount < 10 ? "  " : (itemCount < 100 ? " " : "")).append(" DATA ");
                        for (int row = 0; row < 8; row += 2) {
                            sbLine.append(">");
                            int[] rowColors = charColors[row];
                            sbLine.append(Integer.toHexString(rowColors[1]).toUpperCase());
                            sbLine.append(Integer.toHexString(rowColors[0]).toUpperCase());
                            rowColors = charColors[row + 1];
                            sbLine.append(Integer.toHexString(rowColors[1]).toUpperCase());
                            sbLine.append(Integer.toHexString(rowColors[0]).toUpperCase());
                            if (row < 6) {
                                sbLine.append(",");
                            }
                        }
                        printPaddedLine(bw, sbLine.toString(), includeComments ? (usedChars[i] ? "" : "unused") : null);
                        itemCount++;
                    }
                }
            }
            else {
                int nColorSets = 0;
                for (int ckey : colorSets.keySet()) {
                    int size = colorSets.get(ckey).size();
                    nColorSets += size / 8 + (size % 8 == 0 ? 0 : 1);
                }
                printPaddedLine(bw, "CLRNUM DATA " + nColorSets, includeComments);
                boolean first = true;
                StringBuilder sbLine = new StringBuilder();
                int itemCount = 0;
                for (int ckey : colorSets.keySet()) {
                    int size = colorSets.get(ckey).size();
                    for (int n = 0; n < size / 8 + (size % 8 == 0 ? 0 : 1); n++) {
                        if (itemCount == 0) {
                            if (first) {
                                sbLine.append("CLRSET BYTE ");
                                first = false;
                            }
                            else {
                                sbLine.append("       BYTE ");
                            }
                        }
                        if (itemCount > 0) {
                            sbLine.append(",");
                        }
                        sbLine.append(">");
                        sbLine.append(Globals.toHexString(ckey, 2));
                        itemCount++;
                        if (itemCount > 3) {
                            printPaddedLine(bw, sbLine.toString(), includeComments);
                            sbLine = new StringBuilder();
                            itemCount = 0;
                        }
                    }
                }
                if (sbLine.length() > 0) {
                    printPaddedLine(bw, sbLine.toString(), includeComments);
                }
            }
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Transition Character Pairs (from, to) ", false);
                printPaddedLine(bw, "****************************************", false);
            }
            printPaddedLine(bw, "TCHNUM DATA " + (imax + 1), includeComments);
            for (int i = 0; i <= imax; i++) {
                TransChar transChar = transCharSet[i];
                if (transChar != null) {
                    printPaddedLine(bw,
                            (i == 0 ? "TCHARS" : "      ") + " BYTE >" + Globals.toHexString(transChar.getFromChar(), 2) + ",>" + Globals.toHexString(transChar.getToChar(), 2),
                            !includeComments ? null :
                                    "#" + Globals.toHexString(transChar.getIndex(), 2) +
                                            (colorMode != Magellan.COLOR_MODE_BITMAP ? " color " + Globals.toHexString(transChar.getForeColor(), 1) + "/" + Globals.toHexString(transChar.getBackColor(), 1) : "") +
                                            (transChar.isColorsOK() ? "" : " ERROR")
                    );
                }
                else {
                    printPaddedLine(bw, (i == 0 ? "TCHARS" : "      ") + " BYTE >FF,>FF", !includeComments ? null : "#" + Globals.toHexString(i, 2) + " unused");
                }
            }
            if (includeComments) {
                printPaddedLine(bw, "****************************************", false);
                printPaddedLine(bw, "* Transition Map Data", false);
                printPaddedLine(bw, "****************************************", false);
            }
            for (int m = 0; m < transMaps.size(); m++) {
                StringBuilder sbLine = new StringBuilder();
                int[][] mapToSave = transMaps.get(m);
                if (includeComments) {
                    printPaddedLine(bw, "* == Map #" + m + " == ", false);
                }
                printPaddedLine(bw, "MC" + m + (m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))) + " DATA " + mapdMain.getScreenColor(m), includeComments);
                sbLine.append("MS").append(m).append(m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))).append(" DATA >");
                sbLine.append(Globals.toHexString(mapToSave[0].length, 4));
                sbLine.append(",>").append(Globals.toHexString(mapToSave.length, 4));
                sbLine.append(",>").append(Globals.toHexString(mapToSave[0].length * mapToSave.length, 4));
                printPaddedLine(bw, sbLine.toString(), "Width, Height, Size");
                sbLine.delete(0, sbLine.length());
                if (!compress) {
                    boolean isFirstByte = true;
                    for (int y = 0; y < mapToSave.length; y++) {
                        if (includeComments) {
                            printPaddedLine(bw, "* -- Map Row " + y + " -- ", false);
                        }
                        for (int cl = 0; cl < Math.ceil((double) mapToSave[y].length / 8); cl++) {
                            if (y == 0 && cl == 0) {
                                sbLine.append("MD").append(m).append(m < 10 ? "   " : (m < 100 ? "  " : (m < 1000 ? " " : ""))).append(" DATA ");
                            }
                            else {
                                sbLine.append("       DATA ");
                            }
                            for (int colpos = (cl * 8); colpos < Math.min((cl + 1) * 8, mapToSave[y].length); colpos++) {
                                if (isFirstByte) {
                                    if (colpos > (cl * 8)) {
                                        sbLine.append(",");
                                    }
                                    sbLine.append(">");
                                }
                                sbLine.append(mapToSave[y][colpos] == MapCanvas.NOCHAR ? "00" : Globals.toHexString(mapToSave[y][colpos], 2));
                                isFirstByte = !isFirstByte;
                            }
                            if (!isFirstByte) {
                                sbLine.append("XX"); // If odd, pad with an illegal value
                                isFirstByte = !isFirstByte;
                            }
                            printPaddedLine(bw, sbLine.toString(), includeComments);
                            sbLine.delete(0, sbLine.length());
                        }
                    }
                }
                else if (!compressWords) {
                    // RLE compression (byte)
                    // We assume all characters are < 128. If msb is set, the next byte determines
                    // how many times (2 - 256) the current byte (with msb cleared) should be repeated.
                    // A repeat count of 0 is used as end marker.
                    System.out.println("Uncompressed size: " + (mapToSave.length * mapToSave[0].length));
                    int i = 0;
                    int n = 0;
                    int current;
                    int last = -1;
                    int count = 0;
                    for (int y = 0; y < mapToSave.length; y++) {
                        for (int x = 0; x < mapToSave[0].length; x++) {
                            current = mapToSave[y][x];
                            if (last != -1) {
                                if (current == last && count < 255) {
                                    // Same byte, increment count
                                    count++;
                                }
                                else if (count > 0) {
                                    // End of run of bytes
                                    i = printByte(bw, sbLine, i, last | 128);
                                    i = printByte(bw, sbLine, i, count);
                                    count = 0;
                                    n += 2;
                                }
                                else {
                                    // Different byte
                                    i = printByte(bw, sbLine, i, last);
                                    n++;
                                }
                            }
                            last = current;
                        }
                    }
                    if (count > 0) {
                        i = printByte(bw, sbLine, i, last | 128);
                        i = printByte(bw, sbLine, i, count);
                        n += 2;
                    }
                    else {
                        // Different byte
                        i = printByte(bw, sbLine, i, last);
                        n++;
                    }
                    // End marker
                    i = printByte(bw, sbLine, i, 128);
                    i = printByte(bw, sbLine, i, 0);
                    n += 2;
                    printPaddedLine(bw, sbLine.toString(), false);
                    n += i;
                    System.out.println("Compressed size: " + n);
                }
                else {
                    // RLE compression (word)
                    // We assume all characters are < 128. If msb of the MSB is set, the byte following the
                    // current word determines how many times (2 - 256) the current word (with msb cleared)
                    // should be repeated. A repeat count of 0 is used as end marker.
                    System.out.println("Uncompressed size: " + (mapToSave.length * mapToSave[0].length));
                    int i = 0;
                    int n = 0;
                    int current;
                    int last = -1;
                    int count = 0;
                    for (int y = 0; y < mapToSave.length; y++) {
                        for (int x = 0; x < mapToSave[0].length - 1; x += 2) {
                            current = mapToSave[y][x] << 8 | mapToSave[y][x + 1];
                            if (last != -1) {
                                if (current == last && count < 255) {
                                    // Same word, increment count
                                    count++;
                                }
                                else if (count > 0) {
                                    // End of run of words
                                    i = printByte(bw, sbLine, i, (last & 0xFF00) >> 8 | 128);
                                    i = printByte(bw, sbLine, i, last & 0x00FF);
                                    i = printByte(bw, sbLine, i, count);
                                    count = 0;
                                    n += 3;
                                }
                                else {
                                    // Different byte
                                    i = printByte(bw, sbLine, i, (last & 0xFF00) >> 8);
                                    i = printByte(bw, sbLine, i, last & 0x00FF);
                                    n += 2;
                                }
                            }
                            last = current;
                        }
                    }
                    if (count > 0) {
                        i = printByte(bw, sbLine, i, (last & 0xFF00) >> 8 | 128);
                        i = printByte(bw, sbLine, i, last & 0x00FF);
                        i = printByte(bw, sbLine, i, count);
                        n += 2;
                    }
                    else {
                        // Different byte
                        i = printByte(bw, sbLine, i, (last & 0xFF00) >> 8);
                        i = printByte(bw, sbLine, i, last & 0x00FF);
                        n++;
                    }
                    // End marker
                    i = printByte(bw, sbLine, i, 128);
                    i = printByte(bw, sbLine, i, 0);
                    i = printByte(bw, sbLine, i, 0);
                    n += 3;
                    printPaddedLine(bw, sbLine.toString(), false);
                    n += i;
                    System.out.println("Compressed size: " + n);
                }
            }
            if (frames > 0 || vertical && frames == -1) {
                if (includeComments) {
                    printPaddedLine(bw, "****************************************", false);
                    printPaddedLine(bw, "* Scrolled Character Patterns", false);
                    printPaddedLine(bw, "****************************************", false);
                }
                if (frames > 0) {
                    for (int f = 0; f < frames; f++) {
                        int offset = f * 8 / frames;
                        for (int i = 0; i <= imax; i++) {
                            StringBuilder sbLine = new StringBuilder();
                            if (i == 0) {
                                sbLine.append(vertical ? "V" : "H").append("PFRM").append(f);
                            }
                            else {
                                sbLine.append("      ");
                            }
                            sbLine.append(" DATA ");
                            String hexstr;
                            TransChar transChar = transCharSet[i];
                            if (transChar != null) {
                                int[][] fromGrid = hmCharGrids.get(transChar.getFromChar());
                                int[][] toGrid = hmCharGrids.get(transChar.getToChar());
                                int[][] scrollGrid = new int[8][8];
                                if (vertical) {
                                    int y1 = 0;
                                    for (int y = 8 - offset; y < 8; y++) {
                                        System.arraycopy(toGrid[y], 0, scrollGrid[y1], 0, 8);
                                        y1++;
                                    }
                                    for (int y = 0; y < 8 - offset; y++) {
                                        System.arraycopy(fromGrid[y], 0, scrollGrid[y1], 0, 8);
                                        y1++;
                                    }
                                }
                                else {
                                    int x1 = 0;
                                    for (int x = offset; x < 8; x++) {
                                        for (int y = 0; y < 8; y++) {
                                            scrollGrid[y][x1] = fromGrid[y][x];
                                        }
                                        x1++;
                                    }
                                    for (int x = 0; x < offset; x++) {
                                        for (int y = 0; y < 8; y++) {
                                            scrollGrid[y][x1] = toGrid[y][x];
                                        }
                                        x1++;
                                    }
                                }
                                hexstr = Globals.getByteString(scrollGrid).toUpperCase();
                            }
                            else {
                                hexstr = Globals.BLANKCHAR;
                            }
                            sbLine.append(">").append(hexstr.substring(0, 4)).append(",");
                            sbLine.append(">").append(hexstr.substring(4, 8)).append(",");
                            sbLine.append(">").append(hexstr.substring(8, 12)).append(",");
                            sbLine.append(">").append(hexstr.substring(12, 16));
                            printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + (transChar == null ? " unused" : "") : null);
                        }
                    }
                }
                // Pattern strips
                else {
                    for (int i = 0; i <= imax; i++) {
                        String hexstr;
                        TransChar transChar = transCharSet[i];
                        if (transChar != null) {
                            hexstr =
                                    Globals.getByteString(hmCharGrids.get(transChar.getToChar())).toUpperCase() +
                                            Globals.getByteString(hmCharGrids.get(transChar.getFromChar())).toUpperCase();
                        }
                        else {
                            hexstr = Globals.BLANKCHAR + Globals.BLANKCHAR;
                        }
                        StringBuilder sbLine = new StringBuilder();
                        sbLine.append(i == 0 ? "PSTRIP" : "      ").append(" DATA ");
                        sbLine.append(">").append(hexstr.substring(0, 4)).append(",");
                        sbLine.append(">").append(hexstr.substring(4, 8)).append(",");
                        sbLine.append(">").append(hexstr.substring(8, 12)).append(",");
                        sbLine.append(">").append(hexstr.substring(12, 16));
                        printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + " " + (transChar == null ? "unused" : "- " + Globals.toHexString(transChar.getToChar(), 2)) : null);
                        sbLine = new StringBuilder();
                        sbLine.append("       DATA ");
                        sbLine.append(">").append(hexstr.substring(16, 20)).append(",");
                        sbLine.append(">").append(hexstr.substring(20, 24)).append(",");
                        sbLine.append(">").append(hexstr.substring(24, 28)).append(",");
                        sbLine.append(">").append(hexstr.substring(28, 32));
                        printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + " " + (transChar == null ? " unused" : "- " + Globals.toHexString(transChar.getFromChar(), 2)) : null);
                    }
                }
            }
            if ((frames > 0 || vertical && frames == -1) && colorMode == Magellan.COLOR_MODE_BITMAP) {
                if (includeComments) {
                    printPaddedLine(bw, "****************************************", false);
                    printPaddedLine(bw, "* Scrolled Character Colors", false);
                    printPaddedLine(bw, "****************************************", false);
                }
                if (frames > 0) {
                    for (int f = 0; f < (vertical ? frames : 1); f++) {
                        int offset = f * 8 / frames;
                        for (int i = 0; i <= imax; i++) {
                            StringBuilder sbLine = new StringBuilder();
                            if (i == 0) {
                                sbLine.append(vertical ? "V" : "H").append("CFRM").append(f);
                            }
                            else {
                                sbLine.append("      ");
                            }
                            sbLine.append(" DATA ");
                            String hexstr;
                            TransChar transChar = transCharSet[i];
                            if (transChar != null) {
                                int[][] fromGrid = hmCharColors.get(transChar.getFromChar());
                                int[][] toGrid = hmCharColors.get(transChar.getToChar());
                                hexstr = "";
                                if (vertical) {
                                    for (int y = 8 - offset; y < 8; y++) {
                                        hexstr += Integer.toHexString(toGrid[y][1]).toUpperCase();
                                        hexstr += Integer.toHexString(toGrid[y][0]).toUpperCase();
                                    }
                                    for (int y = 0; y < 8 - offset; y++) {
                                        hexstr += Integer.toHexString(fromGrid[y][1]).toUpperCase();
                                        hexstr += Integer.toHexString(fromGrid[y][0]).toUpperCase();
                                    }
                                }
                                else {
                                    for (int y = 0; y < 8; y++) {
                                        int foreColor = 0;
                                        int backColor = 0;
                                        int fromForeColor = fromGrid[y][Globals.INDEX_CLR_FORE] != 0 ? fromGrid[y][Globals.INDEX_CLR_FORE] : mapdMain.getColorScreen();
                                        int toForeColor = toGrid[y][Globals.INDEX_CLR_FORE] != 0 ? toGrid[y][Globals.INDEX_CLR_FORE] : mapdMain.getColorScreen();
                                        if (fromForeColor == toForeColor) {
                                            foreColor = fromForeColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(transChar.getFromChar())[y], Globals.INDEX_CLR_FORE)) {
                                            foreColor = toForeColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(transChar.getToChar())[y], Globals.INDEX_CLR_FORE)) {
                                            foreColor = fromForeColor;
                                        }
                                        int fromBackColor = fromGrid[y][Globals.INDEX_CLR_BACK] != 0 ? fromGrid[y][Globals.INDEX_CLR_BACK] : mapdMain.getColorScreen();
                                        int toBackColor = toGrid[y][Globals.INDEX_CLR_BACK] != 0 ? toGrid[y][Globals.INDEX_CLR_BACK] : mapdMain.getColorScreen();
                                        if (fromBackColor == toBackColor) {
                                            backColor = fromBackColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(transChar.getFromChar())[y], Globals.INDEX_CLR_BACK)) {
                                            backColor = toBackColor;
                                        }
                                        else if (!Globals.arrayContains(hmCharGrids.get(transChar.getToChar())[y], Globals.INDEX_CLR_BACK)) {
                                            backColor = fromBackColor;
                                        }
                                        hexstr += Integer.toHexString(foreColor).toUpperCase();
                                        hexstr += Integer.toHexString(backColor).toUpperCase();
                                    }
                                }
                            }
                            else {
                                hexstr = Globals.BLANKCHAR;
                            }
                            sbLine.append(">").append(hexstr.substring(0, 4)).append(",");
                            sbLine.append(">").append(hexstr.substring(4, 8)).append(",");
                            sbLine.append(">").append(hexstr.substring(8, 12)).append(",");
                            sbLine.append(">").append(hexstr.substring(12, 16));
                            printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + (transChar == null ? " unused" : "") : null);
                        }
                    }
                }
                // Color strips
                else {
                    for (int i = 0; i <= imax; i++) {
                        String hexstr = "";
                        TransChar transChar = transCharSet[i];
                        if (transChar != null) {
                            int[][] toColors = hmCharColors.get(transChar.getToChar());
                            for (int row = 0; row < 8; row++) {
                                hexstr += Integer.toHexString(toColors[row][1]) + Integer.toHexString(toColors[row][0]);
                            }
                            int[][] fromColors = hmCharColors.get(transChar.getFromChar());
                            for (int row = 0; row < 8; row++) {
                                hexstr += Integer.toHexString(fromColors[row][1]) + Integer.toHexString(fromColors[row][0]);
                            }
                        }
                        else {
                            hexstr = Globals.BLANKCHAR + Globals.BLANKCHAR;
                        }
                        hexstr = hexstr.toUpperCase();
                        StringBuilder sbLine = new StringBuilder();
                        sbLine.append(i == 0 ? "CSTRIP" : "      ").append(" DATA ");
                        sbLine.append(">").append(hexstr.substring(0, 4)).append(",");
                        sbLine.append(">").append(hexstr.substring(4, 8)).append(",");
                        sbLine.append(">").append(hexstr.substring(8, 12)).append(",");
                        sbLine.append(">").append(hexstr.substring(12, 16));
                        printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + " " + (transChar == null ? "unused" : "- " + Globals.toHexString(transChar.getToChar(), 2)) : null);
                        sbLine = new StringBuilder();
                        sbLine.append("       DATA ");
                        sbLine.append(">").append(hexstr.substring(16, 20)).append(",");
                        sbLine.append(">").append(hexstr.substring(20, 24)).append(",");
                        sbLine.append(">").append(hexstr.substring(24, 28)).append(",");
                        sbLine.append(">").append(hexstr.substring(28, 32));
                        printPaddedLine(bw, sbLine.toString(), includeComments ? "#" + Globals.toHexString(i, 2) + " " + (transChar == null ? " unused" : "- " + Globals.toHexString(transChar.getFromChar(), 2)) : null);
                    }
                }
            }
            bw.flush();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        } finally {
            if (bw != null) {
                try {
                    bw.close();
                } catch (IOException e) {
                    e.printStackTrace(System.out);
                }
            }
        }
    }

    protected int printByte(BufferedWriter bw, StringBuilder sbLine, int i, int b) throws IOException {
        sbLine.append(i == 0 ? "       BYTE " : ",").append(">").append(Globals.toHexString(b, 2));
        i++;
        if (i == 8) {
            printPaddedLine(bw, sbLine.toString(), false);
            sbLine.delete(0, sbLine.length());
            i = 0;
        }
        return i;
    }

    protected void writeBinaryFile(File mapDataFile, byte chunkFlags, int startChar, int endChar, boolean currMapOnly) {
        try {
            // store working map first
            mapdMain.storeCurrentMap();
            // get file output buffer
            FileOutputStream fos = new FileOutputStream(mapDataFile);

            // write File Header
            fos.write(BIN_HEADER_MAG);
            fos.write(BIN_HEADER_VER);
            fos.write(chunkFlags);
            byte mapCount = (byte) (currMapOnly ? 1 : mapdMain.getMapCount());
            fos.write(mapCount);

            // write Colorset Chunk (if present)
            if ((chunkFlags & BIN_CHUNK_COLORS) == BIN_CHUNK_COLORS) {
                if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                    byte charCount = (byte) ((endChar - startChar) + 1);
                    fos.write(charCount);
                    for (int bc = startChar; bc <= endChar; bc++) {
                        fos.write(getColorBytes(bc));
                    }
                }
                else {
                    for (int i = 0; i < clrSets.length; i++) {
                        fos.write((byte) (clrSets[i][Globals.INDEX_CLR_FORE] << 4 | clrSets[i][Globals.INDEX_CLR_BACK]));
                    }
                }
            }

            // write Character Chunk (if present)
            if ((chunkFlags & BIN_CHUNK_CHARS) == BIN_CHUNK_CHARS) {
                byte charCount = (byte) ((endChar - startChar) + 1);
                fos.write(charCount);
                for (int bc = startChar; bc <= endChar; bc++) {
                    fos.write(getCharBytes(bc));
                }
            }

            // write Maps
            for (int m = 0; m < mapdMain.getMapCount(); m++) {
                if (!currMapOnly || m == mapdMain.getCurrentMapId()) {
                    int[][] mapToSave = mapdMain.getMapData(m);
                    int mapCols = mapToSave[0].length;
                    int mapRows = mapToSave.length;
                    int mapSize = mapCols * mapRows;
                    int mapScreenColor = mapdMain.getScreenColor(m);

                    // write Map Header
                    //   reserved bytes for Magellan use
                    fos.write(BIN_MAP_HEADER_RESERVED1);
                    fos.write(BIN_MAP_HEADER_RESERVED2);
                    fos.write(BIN_MAP_HEADER_RESERVED3);
                    fos.write(BIN_MAP_HEADER_RESERVED4);
                    //   map size as a series of bytes
                    int mapSizeChunk = mapSize;
                    for (int i = 0; i < 8; i++) {
                        if (mapSizeChunk > 255) {
                            fos.write(255);
                            mapSizeChunk -= 255;
                        }
                        else if (mapSizeChunk > 0) {
                            fos.write((byte) mapSizeChunk);
                            mapSizeChunk = 0;
                        }
                        else {
                            fos.write(0);
                        }
                    }
                    //   map columns as a byte pair
                    if (mapCols > 255) {
                        fos.write(255);
                        fos.write(mapCols - 255);
                    }
                    else {
                        fos.write(mapCols);
                        fos.write(0);
                    }
                    //   map rows as a byte pair
                    if (mapRows > 255) {
                        fos.write(255);
                        fos.write(mapRows - 255);
                    }
                    else {
                        fos.write(mapRows);
                        fos.write(0);
                    }
                    //   map screen color
                    fos.write(mapScreenColor);

                    // write Map Data
                    for (int y = 0; y < mapRows; y++) {
                        for (int x = 0; x < mapCols; x++) {
                            fos.write(mapToSave[y][x]);
                        }
                    }

                    // write Sprite Chunk (if present)
                    if ((chunkFlags & BIN_CHUNK_SPRITES) == BIN_CHUNK_SPRITES) {
                        // not yet implemented
                    }
                }
            }

            fos.flush();
            fos.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void writeBinaryMap(File mapDataFile) {
        try {
            // store working map first
            mapdMain.storeCurrentMap();
            // get file output buffer
            FileOutputStream fos = new FileOutputStream(mapDataFile);
            // write map
            int[][] mapToSave = mapdMain.getMapData(mapdMain.getCurrentMapId());
            for (int[] mapRow : mapToSave) {
                for (int mapChar : mapRow) {
                    fos.write(mapChar);
                }
            }
            fos.flush();
            fos.close();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected void printPaddedLine(BufferedWriter bw, String str, boolean isCommand, int padlen, String comment)
            throws IOException {
        if (str.length() < padlen) {
            StringBuffer sbOut = new StringBuffer();
            sbOut.append(str);
            for (int i = str.length(); i < padlen; i++) {
                if ((i == (padlen - 1)) && isCommand) {
                    sbOut.append(";");
                }
                else {
                    sbOut.append(" ");
                }
            }
            if (comment != null) {
                sbOut.append(" ").append(comment);
            }
            bw.write(sbOut.toString());
        }
        else {
            bw.write(str);
        }
        bw.newLine();
    }

    protected void printPaddedLine(BufferedWriter bw, String str, boolean isCommand) throws IOException {
        printPaddedLine(bw, str, isCommand, ASM_LINELEN, null);
    }

    protected void printPaddedLine(BufferedWriter bw, String str, String comment) throws IOException {
        printPaddedLine(bw, str, comment != null, ASM_LINELEN, comment);
    }

    protected void readCharImageMono(BufferedImage buffImg) {
        // get character glyphs
        int rowOffset = 0;
        int colOffset = 0;
        for (int charNum = TIGlobals.MINCHAR; charNum <= TIGlobals.MAXCHAR; charNum++) {
            int[][] newCharArray = new int[8][8];
            if (hmCharGrids.containsKey(charNum)) {
                hmCharGrids.remove(charNum);
            }
            for (int y = 0; y < 8; y++) {
                for (int x = 0; x < 8; x++) {
                    newCharArray[y][x] = (buffImg.getRGB((colOffset * 8) + x, (rowOffset * 8) + y) != -1 ? 1 : 0);
                }
            }
            hmCharGrids.put(charNum, newCharArray);
            colOffset++;
            if (colOffset >= 8) {
                colOffset = 0;
                rowOffset++;
            }
        }
    }

    protected void readCharImageColor(BufferedImage buffImg) {
        // get colorsets
        for (int cs = 0; cs < 32; cs++) {
            int colorFore = -1;
            int colorBack = -1;
            int testY = 0;
            int testX = 0;
            while (testY < 8 && (colorFore == -1 || colorBack == -1)) {
                int pixelColor = buffImg.getRGB(testX, (cs * 8) + testY);
                int tiColor = getTIColorForPixel(pixelColor);
                if ((tiColor > 0) && (colorFore == -1)) {
                    colorFore = tiColor;
                }
                else if ((tiColor > 0) && (tiColor != colorFore)) {
                    colorBack = tiColor;
                }
                testX++;
                if (testX >= 64) {
                    testX = 0;
                    testY++;
                }
            }
            if (colorFore == -1) {
                colorFore = 1;
            }
            if (colorBack == -1) {
                colorBack = 0;
            }
            clrSets[cs][Globals.INDEX_CLR_FORE] = colorFore;
            clrSets[cs][Globals.INDEX_CLR_BACK] = colorBack;
        }
        // Global ECM palette
        ECMPalette ecmGlobalPalette = null;
        if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
            ECMPalette basePalette = ecmPalettes[0];
            int ecmColorIndex = 0;
            for (int y = 0; y < buffImg.getHeight(); y++) {
                for (int x = 0; x < buffImg.getWidth(); x++) {
                    int color = buffImg.getRGB(x, y);
                    boolean found = false;
                    for (int i = 0; i < ecmColorIndex && !found; i++) {
                        if (basePalette.getColor(i).getRGB() == color) {
                            found = true;
                        }
                    }
                    if (!found && ecmColorIndex < basePalette.getSize()) {
                        basePalette.setColor(ecmColorIndex++, new Color(color));
                    }
                }
            }
            if (ecmColorIndex < basePalette.getSize()) {
                for (int i = ecmColorIndex; i < basePalette.getSize(); i++) {
                    basePalette.setColor(i, Color.WHITE);
                }
                basePalette.sort();
                ecmGlobalPalette = basePalette;
            }
        }
        // get character glyphs
        int rowOffset = 0;
        int colOffset = 0;
        int cSet = 0;
        int ecmPaletteIndex = 1;
        for (int charNum = TIGlobals.MINCHAR; charNum <= TIGlobals.MAXCHAR; charNum++) {
            int[][] newCharArray = new int[8][8];
            int[][] newColorArray = new int[8][2];
            // ECM palette for character
            int[] ecmColors = colorMode == Magellan.COLOR_MODE_ECM_2 ? new int[4] : (colorMode == Magellan.COLOR_MODE_ECM_3 ? new int[8] : null);
            int ecmColorIndex = 0;
            if (ecmGlobalPalette == null && (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3)) {
                for (int y = 0; y < 8; y++) {
                    for (int x = 0; x < 8; x++) {
                        int color = buffImg.getRGB((colOffset * 8) + x, (rowOffset * 8) + y);
                        boolean found = false;
                        for (int i = 0; i < Math.min(ecmColorIndex, ecmColors.length) && !found; i++) {
                            if (ecmColors[i] == color) {
                                found = true;
                            }
                        }
                        if (!found) {
                            if (ecmColorIndex < ecmColors.length) {
                                ecmColors[ecmColorIndex++] = color;
                            }
                        }
                    }
                }
                for (int i = ecmColorIndex; i < ecmColors.length; i++) {
                    ecmColors[i] = Color.WHITE.getRGB();
                }
                Arrays.sort(ecmColors);
            }
            // ...
            if (hmCharGrids.containsKey(charNum)) {
                hmCharGrids.remove(charNum);
            }
            for (int y = 0; y < 8; y++) {
                int[] newColors = newColorArray[y];
                newColors[0] = -1;
                newColors[1] = -1;
                for (int x = 0; x < 8; x++) {
                    int color = buffImg.getRGB((colOffset * 8) + x, (rowOffset * 8) + y) | 0xff000000;
                    if (colorMode == Magellan.COLOR_MODE_GRAPHICS_1) {
                        newCharArray[y][x] = (color == TIGlobals.TI_PALETTE_OPAQUE[clrSets[cSet][Globals.INDEX_CLR_FORE]].getRGB() ? 1 : 0);
                    }
                    else if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                        int tiColor = getTIColorForPixel(color);
                        if (newColors[0] == -1) {
                            newColors[0] = tiColor;
                        }
                        else if (newColors[1] == -1 && newColors[0] != tiColor) {
                            newColors[1] = tiColor;
                        }
                        newCharArray[y][x] = tiColor == newColors[0] ? 0 : 1;
                    }
                    else {
                        boolean found = false;
                        if (ecmGlobalPalette == null) {
                            for (int i = 0; i < ecmColors.length && !found; i++) {
                                if (ecmColors[i] == color) {
                                    newCharArray[y][x] = i;
                                    found = true;
                                }
                            }
                        }
                        else {
                            for (int i = 0; i < ecmGlobalPalette.getSize() && !found; i++) {
                                if (ecmGlobalPalette.getColor(i).getRGB() == color) {
                                    newCharArray[y][x] = i;
                                    found = true;
                                }
                            }
                        }
                    }
                }
            }
            hmCharGrids.put(charNum, newCharArray);
            if (colorMode == Magellan.COLOR_MODE_BITMAP) {
                hmCharColors.put(charNum, newColorArray);
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                if (ecmGlobalPalette == null) {
                    // Match char palette to existing palettes
                    boolean found = false;
                    for (int i = 0; i < Math.min(ecmPaletteIndex, ecmPalettes.length); i++) {
                        if (ecmPalettes[i].startsWith(ecmColors, ecmColorIndex)) {
                            ecmCharPalettes[charNum] = ecmPalettes[i];
                            found = true;
                        }
                    }
                    if (!found) {
                        if (ecmPaletteIndex < ecmPalettes.length - 1) {
                            ecmPalettes[ecmPaletteIndex].setIntColors(ecmColors);
                            ecmCharPalettes[charNum] = ecmPalettes[ecmPaletteIndex];
                            ecmPaletteIndex++;
                        }
                        else {
                            ecmCharPalettes[charNum] = ecmPalettes[ecmPaletteIndex];
                        }
                    }
                }
                else {
                    ecmCharPalettes[charNum] = ecmGlobalPalette;
                }
            }
            colOffset++;
            if (colOffset >= 8) {
                colOffset = 0;
                rowOffset++;
                cSet++;
            }
        }
    }

    protected void writeCharImage(File imageOut, int tileCols, boolean isColor) {
        int charCount = hmCharGrids.size();
        int tileRows = (int) Math.ceil(charCount / tileCols);
        BufferedImage bufferCharImage = new BufferedImage(tileCols * 8, tileRows * 8, BufferedImage.TYPE_INT_ARGB);
        if (bufferCharImage == null) {
            bufferCharImage = new BufferedImage(tileCols * 8, tileRows * 8, BufferedImage.TYPE_INT_ARGB);
        }
        try {
            Thread.sleep(500);
        } catch (InterruptedException ie) {
            System.out.println(ie.getMessage());
        }
        if (bufferCharImage == null) {
            System.out.println("Unable to initialize BufferedImage");
        }
        else {
            int[][] charGrid;
            int[][] charColors;
            int drawRow = 0;
            int drawCol = 0;
            int setNum = 0;
            boolean shouldDraw;
            try {
                Graphics2D gf = (Graphics2D) (bufferCharImage.getGraphics());
                if (isColor) {
                    gf.setComposite(AlphaComposite.Clear);
                    gf.setColor(TIGlobals.TI_COLOR_TRANSPARENT);
                }
                else {
                    gf.setComposite(AlphaComposite.SrcOver);
                    gf.setColor(TIGlobals.TI_COLOR_WHITE);
                }
                gf.fillRect(0, 0, bufferCharImage.getWidth(), bufferCharImage.getHeight());
                gf.setColor(TIGlobals.TI_COLOR_BLACK);
                for (int ch = 0; ch < charCount; ch++) {
                    gf.setComposite(AlphaComposite.SrcOver);
                    if (hmCharGrids.containsKey(ch)) {
                        charGrid = hmCharGrids.get(ch);
                        charColors = colorMode == Magellan.COLOR_MODE_BITMAP ? hmCharColors.get(ch) : null;
                        for (int y = 0; y < charGrid.length; y++) {
                            for (int x = 0; x < charGrid[y].length; x++) {
                                shouldDraw = true;
                                if (colorMode == Magellan.COLOR_MODE_GRAPHICS_1 || colorMode == Magellan.COLOR_MODE_BITMAP) {
                                    if (charGrid[y][x] == 1) {
                                        if (isColor) {
                                            if (charColors == null) {
                                                if (clrSets[setNum][Globals.INDEX_CLR_FORE] == 0) {
                                                    shouldDraw = false;
                                                }
                                                gf.setColor(TIGlobals.TI_PALETTE[clrSets[setNum][Globals.INDEX_CLR_FORE]]);
                                            }
                                            else {
                                                shouldDraw = charColors[y][1] != 0;
                                                gf.setColor(TIGlobals.TI_PALETTE[charColors[y][1]]);
                                            }
                                        }
                                        else {
                                            gf.setColor(TIGlobals.TI_COLOR_BLACK);
                                        }
                                    }
                                    else {
                                        if (isColor) {
                                            if (charColors == null) {
                                                if (clrSets[setNum][Globals.INDEX_CLR_BACK] == 0) {
                                                    shouldDraw = false;
                                                }
                                                gf.setColor(TIGlobals.TI_PALETTE[clrSets[setNum][Globals.INDEX_CLR_BACK]]);
                                            }
                                            else {
                                                shouldDraw = charColors[y][0] != 0;
                                                gf.setColor(TIGlobals.TI_PALETTE[charColors[y][0]]);
                                            }
                                        }
                                        else {
                                            gf.setColor(TIGlobals.TI_COLOR_WHITE);
                                        }
                                    }
                                }
                                else {
                                    if (isColor) {
                                        gf.setColor(ecmCharPalettes[ch].getColor(charGrid[y][x]));
                                    }
                                    else {
                                        gf.setColor(charGrid[y][x] != 0 ? TIGlobals.TI_COLOR_BLACK : TIGlobals.TI_COLOR_WHITE);
                                    }
                                }
                                if (shouldDraw) {
                                    gf.fillRect((drawCol * 8) + x, (drawRow * 8) + y, 1, 1);
                                }
                            }
                        }
                    }
                    drawCol++;
                    if (drawCol >= tileCols) {
                        drawCol = 0;
                        drawRow++;
                        setNum++;
                    }
                }
                gf.dispose();
                ImageIO.write(bufferCharImage, "png", imageOut);
            } catch (Exception e) {
                e.printStackTrace(System.out);
            }
        }
    }

    protected void writeMapImage(File imageOut) {
        try {
            ImageIO.write(mapdMain.getBuffer(), "png", imageOut);
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    protected int getTIColorForPixel(int pixelRBG) {
        for (int c = 0; c < TIGlobals.TI_PALETTE_OPAQUE.length; c++) {
            if (pixelRBG == TIGlobals.TI_PALETTE_OPAQUE[c].getRGB()) {
                return c;
            }
        }
        return 0;
    }

    protected byte[] getColorBytes(int charnum) {
        byte[] colorbytes = new byte[8];
        int[][] colorarray = hmCharColors.get(charnum);
        for (int y = 0; y < 8; y++) {
            byte background = (byte) colorarray[y][0];
            byte foreground = (byte) (colorarray[y][1] << 4);
            colorbytes[y] = (byte) (background | foreground);
        }
        return colorbytes;
    }

    protected byte[] getCharBytes(int charnum) {
        byte[] charbytes = new byte[8];
        int[][] chararray = hmCharGrids.get(charnum);
        if (chararray != null) {
            int bcount = 0;
            byte byteval = (byte) 0;
            int bytepos = 0;
            boolean goHigh = true;
            for (int y = 0; y < 8; y++) {
                bytepos = 0;
                for (int x = 0; x < 8; x++) {
                    if (chararray[y][x] > 0) {
                        byteval = (byte) (byteval | (bytepos == 0 ? 8 : (bytepos == 1 ? 4 : (bytepos == 2 ? 2 : 1))));
                    }
                    bytepos++;
                    if (bytepos > 3) {
                        charbytes[bcount] = (byte) (charbytes[bcount] | (goHigh ? byteval << 4 : byteval));
                        if (!goHigh) {
                            bcount++;
                        }
                        goHigh = !goHigh;
                        byteval = (byte) 0;
                        bytepos = 0;
                    }
                }
            }
        }
        else {
            for (int i = 0; i < charbytes.length; i++) {
                charbytes[i] = (byte) 0;
            }
        }
        return charbytes;
    }
}
